/*------------------------------------------------------------------------*/
/* This file provides an application that runs the COMPSAT library on
 * multiple input files in the DIMACS format.  It produces a one line
 * output for all files and finally a summary.
 */
#include "compsat.h"
#include "config.h"

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>

/*------------------------------------------------------------------------*/

static char *
common_prefix (char **files)
{
  const char *p, *q;
  int i, len, tmp;
  char *res;

  if (files[0] && files[1])
    {
      len = strlen (files[0]);
      i = 1;

      while ((p = files[i++]))
	{
	  for (q = files[0]; *p && *p == *q; p++, q++)
	    ;

	  if (len > (tmp = q - files[0]))
	    len = tmp;
	}
    }
  else
    len = 0;

  res = (char *) malloc (len + 1);
  for (i = 0; i < len; i++)
    res[i] = files[0][i];
  res[len] = 0;

  return res;
}

/*------------------------------------------------------------------------*/

static void
reverse_string (char *s)
{
  char *p, *q, tmp;

  for (p = s; *p; p++)
    ;

  p--;

  for (q = s; q < p; p--, q++)
    {
      tmp = *p;
      *p = *q;
      *q = tmp;
    }
}

/*------------------------------------------------------------------------*/

static void
reverse_files (char **files)
{
  char **s;

  for (s = files; *s; s++)
    reverse_string (*s);
}

/*------------------------------------------------------------------------*/

static char *
common_suffix (char **files)
{
  char *res;

  reverse_files (files);
  res = common_prefix (files);
  reverse_files (files);
  reverse_string (res);

  return res;
}

/*------------------------------------------------------------------------*/

#define USAGE_WITHOUT_CHECK \
"mcompsat [<option> ...] <file> ...\n" \
"\n" \
"where <option> is one of the following\n" \
"\n" \
"  -h             print this command line summary\n" \
"  --version      print version information\n" \
"  --copyright    print copyright information\n" \
"\n" \
"  -d[ ]<num>     max decisions per instance (default -1 = unrestricted)\n" \
"  -m[ ]<num>     max mega bytes per instance (default -1 = unrestricted)\n" \
"  -t[ ]<num>     max seconds per instance (default -1 = unrestricted)\n"

/*------------------------------------------------------------------------*/

#ifdef COMPSAT_CHECK
#define USAGE \
USAGE_WITHOUT_CHECK \
"\n" \
"  --check        enable expensive checking code\n"
#else
#define USAGE USAGE_WITHOUT_CHECK
#endif

/*------------------------------------------------------------------------*/

static void
usage (CompSat * compsat)
{
  printf (USAGE);
}

/*------------------------------------------------------------------------*/

static void
line (void)
{
  int i;

  for (i = 0; i < 78; i++)
    fputc ('-', stdout);

  fputc ('\n', stdout);
}

/*------------------------------------------------------------------------*/

static void
print_name (const char *name, int prefix_length, int suffix_length)
{
  int i, l;

  l = strlen (name);

  for (i = prefix_length; i < l - suffix_length; i++)
    fputc (name[i], stdout);
}

/*------------------------------------------------------------------------*/

int
main (int argc, char **argv)
{
  int count_abort, count_err, count_sat, count_unsat, count_files;
  int i, size_files, processed, found, done, res;
  int prefix_length, suffix_length;
  char **files, *prefix, *suffix;
  double max_mb, sum_sec, mb, sec;
  COMPSAT_ArgvIterator *it;
  const char *name, *err;
  CompSat *compsat;

  count_files = 0;
  size_files = 16;
  files = (char **) malloc (sizeof (char *) * size_files);
  err = 0;

  compsat = COMPSAT_new ();
  it = COMPSAT_new_ArgvIterator (compsat, argc, argv);
  while (!err && COMPSAT_has_more_elements_ArgvIterator (it, &err))
    {
      if (it->opt[0] != '-')
	{
	  if (count_files >= size_files - 1)
	    files =
	      (char **) realloc (files, sizeof (char **) * (size_files *= 2));

	  files[count_files++] = strdup (it->opt);
	}
    }
  if (err && count_files)
    {
      COMPSAT_option (compsat, "--divert", 0);
      COMPSAT_report_invalid_option (compsat, it, err);
      suffix_length = prefix_length = 0;
    }
  else
    {
      files[count_files] = 0;
      printf ("will process %d filess\n", count_files);
      prefix = common_prefix (files);
      prefix_length = strlen (prefix);
      printf ("common prefix %s\n", prefix);
      suffix = common_suffix (files);
      suffix_length = strlen (suffix);
      printf ("common suffix %s\n", suffix);
      free (prefix);
      free (suffix);
      line ();
    }
  COMPSAT_delete_ArgvIterator (it);
  COMPSAT_delete (compsat);
  for (i = 0; i < count_files; i++)
    free (files[i]);
  free (files);

  count_abort = count_err = count_sat = count_unsat = 0;
  done = found = processed = 0;
  sum_sec = max_mb = 0;

  while (!err && !done)
    {
      compsat = COMPSAT_new ();
      found = 0;
      name = 0;
      it = COMPSAT_new_ArgvIterator (compsat, argc, argv);
      while (!err && COMPSAT_has_more_elements_ArgvIterator (it, &err))
	{
	  if (!strcmp (it->opt, "-h"))
	    {
	      usage (compsat);
	      exit (0);
	    }
	  else if (!strcmp (it->opt, "--version"))
	    {
	      printf ("%s\n", COMPSAT_version ());
	      exit (0);
	    }
	  else if (!strcmp (it->opt, "--copyright"))
	    {
	      printf ("%s\n", COMPSAT_copyright ());
	      exit (0);
	    }
	  else if (!strcmp (it->opt, "-d") ||
#ifdef COMPSAT_CHECK
		   !strcmp (it->opt, "--check") ||
#endif
		   !strcmp (it->opt, "-m") || !strcmp (it->opt, "-t"))
	    {
	      err = COMPSAT_option (compsat, it->opt, it->arg);
	    }
	  else if (it->opt[0] != '-')
	    {
	      if (found++ == processed)
		name = it->opt;
	    }
	  else
	    err = "invalid command line option (for mcompsat)";
	}

      if (!err)
	{
	  if (name)
	    {
	      print_name (name, prefix_length, suffix_length);
	      fflush (stdout);
	      if (COMPSAT_read (compsat, name))
		{
		  printf (" ERROR");
		  count_err++;
		}
	      else
		{
		  res = COMPSAT_sat (compsat);
		  if (res == COMPSAT_SATISFIABLE)
		    count_sat++;
		  else if (res == COMPSAT_UNSATISFIABLE)
		    count_unsat++;
		  else
		    count_abort++;
		  printf (" %s", COMPSAT_res2str (res));
		}

	      mb = COMPSAT_bytes (compsat);
	      mb /= 1 << 20;
	      if (mb > max_mb)
		max_mb = mb;
	      sec = COMPSAT_seconds (compsat);
	      sum_sec += sec;
	      printf (" %.1f MB %.2f sec\n", mb, sec);
	      fflush (stdout);
	      processed++;
	    }
	  else if (processed)
	    done = 1;
	  else
	    {
	      fprintf (stderr, "*** no file specified on command line\n");
	      exit (1);
	    }
	}
      else
	{
	  COMPSAT_option (compsat, "--divert", 0);
	  COMPSAT_report_invalid_option (compsat, it, err);
	}

      COMPSAT_delete_ArgvIterator (it);
      COMPSAT_delete (compsat);
    }

  if (!err)
    {
      line ();
      printf
	("%d files, %d error, %d aborted, %d sat, %d unsat, %.1f max MB, %.2f sec\n",
	 count_files, count_err, count_abort, count_sat, count_unsat, max_mb,
	 sum_sec);
    }

  return err ? 1 : 0;
}
